@@ -1,8 +1,7 @@
-use crate::{
-    fs::{OpenOptions, Permissions},
-    fs_utf8::{from_utf8, to_utf8, DirBuilder, File, Metadata, ReadDir},
-};
+use crate::fs::{OpenOptions, Permissions};
+use crate::fs_utf8::{from_utf8, to_utf8, DirBuilder, File, Metadata, ReadDir};
 use async_std::{fs, io};
+use camino::{Utf8Path, Utf8PathBuf};
 use cap_primitives::{ambient_authority, AmbientAuthority};
 #[cfg(not(windows))]
 use io_lifetimes::{AsFd, BorrowedFd, FromFd, IntoFd, OwnedFd};
@@ -31,6 +30,7 @@ use {
 /// since absolute paths don't interoperate well with the capability model.
 ///
 /// [functions in `async_std::fs`]: https://docs.rs/async-std/latest/async_std/fs/index.html#functions
+#[derive(Clone)]
 pub struct Dir {
     cap_std: crate::fs::Dir,
 }
@@ -62,9 +62,9 @@ impl Dir {
     /// This corresponds to [`async_std::fs::File::open`], but only accesses
     /// paths relative to `self`.
     #[inline]
-    pub fn open<P: AsRef<str>>(&self, path: P) -> io::Result<File> {
+    pub async fn open<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<File> {
         let path = from_utf8(path)?;
-        self.cap_std.open(path).map(File::from_cap_std)
+        self.cap_std.open(path).await.map(File::from_cap_std)
     }
 
     /// Opens a file at `path` with the options specified by `options`.
@@ -74,26 +74,33 @@ impl Dir {
     /// Instead of being a method on `OpenOptions`, this is a method on `Dir`,
     /// and it only accesses paths relative to `self`.
     #[inline]
-    pub fn open_with<P: AsRef<str>>(&self, path: P, options: &OpenOptions) -> io::Result<File> {
+    pub async fn open_with<P: AsRef<Utf8Path>>(
+        &self,
+        path: P,
+        options: &OpenOptions,
+    ) -> io::Result<File> {
         let path = from_utf8(path)?;
         self.cap_std
             .open_with(path, options)
+            .await
             .map(File::from_cap_std)
     }
 
     /// Attempts to open a directory.
     #[inline]
-    pub fn open_dir<P: AsRef<str>>(&self, path: P) -> io::Result<Self> {
+    pub async fn open_dir<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<Self> {
         let path = from_utf8(path)?;
-        self.cap_std.open_dir(path).map(Self::from_cap_std)
+        self.cap_std.open_dir(path).await.map(Self::from_cap_std)
     }
 
     /// Creates a new, empty directory at the provided path.
     ///
     /// This corresponds to [`async_std::fs::create_dir`], but only accesses
     /// paths relative to `self`.
+    ///
+    /// TODO: async: fix this when we fix https://github.com/bytecodealliance/cap-std/issues/51
     #[inline]
-    pub fn create_dir<P: AsRef<str>>(&self, path: P) -> io::Result<()> {
+    pub fn create_dir<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<()> {
         let path = from_utf8(path)?;
         self.cap_std.create_dir(path)
     }
@@ -103,8 +110,10 @@ impl Dir {
     ///
     /// This corresponds to [`async_std::fs::create_dir_all`], but only
     /// accesses paths relative to `self`.
+    ///
+    /// TODO: async: fix this when we fix https://github.com/bytecodealliance/cap-std/issues/51
     #[inline]
-    pub fn create_dir_all<P: AsRef<str>>(&self, path: P) -> io::Result<()> {
+    pub fn create_dir_all<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<()> {
         let path = from_utf8(path)?;
         self.cap_std.create_dir_all(path)
     }
@@ -113,8 +122,10 @@ impl Dir {
     /// builder.
     ///
     /// This corresponds to [`async_std::fs::DirBuilder::create`].
+    ///
+    /// TODO: async: fix this when we fix https://github.com/bytecodealliance/cap-std/issues/51
     #[inline]
-    pub fn create_dir_with<P: AsRef<str>>(
+    pub fn create_dir_with<P: AsRef<Utf8Path>>(
         &self,
         path: P,
         dir_builder: &DirBuilder,
@@ -128,9 +139,9 @@ impl Dir {
     /// This corresponds to [`async_std::fs::File::create`], but only accesses
     /// paths relative to `self`.
     #[inline]
-    pub fn create<P: AsRef<str>>(&self, path: P) -> io::Result<File> {
+    pub async fn create<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<File> {
         let path = from_utf8(path)?;
-        self.cap_std.create(path).map(File::from_cap_std)
+        self.cap_std.create(path).await.map(File::from_cap_std)
     }
 
     /// Returns the canonical form of a path with all intermediate components
@@ -140,9 +151,9 @@ impl Dir {
     /// returning an absolute path, returns a path relative to the
     /// directory represented by `self`.
     #[inline]
-    pub fn canonicalize<P: AsRef<str>>(&self, path: P) -> io::Result<String> {
+    pub async fn canonicalize<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<Utf8PathBuf> {
         let path = from_utf8(path)?;
-        self.cap_std.canonicalize(path).and_then(to_utf8)
+        self.cap_std.canonicalize(path).await.and_then(to_utf8)
     }
 
     /// Copies the contents of one file to another. This function will also
@@ -152,7 +163,7 @@ impl Dir {
     /// This corresponds to [`async_std::fs::copy`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub async fn copy<P: AsRef<str>, Q: AsRef<str>>(
+    pub async fn copy<P: AsRef<Utf8Path>, Q: AsRef<Utf8Path>>(
         &self,
         from: P,
         to_dir: &Self,
@@ -168,7 +179,7 @@ impl Dir {
     /// This corresponds to [`async_std::fs::hard_link`], but only accesses
     /// paths relative to `self`.
     #[inline]
-    pub fn hard_link<P: AsRef<str>, Q: AsRef<str>>(
+    pub async fn hard_link<P: AsRef<Utf8Path>, Q: AsRef<Utf8Path>>(
         &self,
         src: P,
         dst_dir: &Self,
@@ -176,7 +187,7 @@ impl Dir {
     ) -> io::Result<()> {
         let src = from_utf8(src)?;
         let dst = from_utf8(dst)?;
-        self.cap_std.hard_link(src, &dst_dir.cap_std, dst)
+        self.cap_std.hard_link(src, &dst_dir.cap_std, dst).await
     }
 
     /// Given a path, query the file system to get information about a file,
@@ -185,15 +196,15 @@ impl Dir {
     /// This corresponds to [`async_std::fs::metadata`], but only accesses
     /// paths relative to `self`.
     #[inline]
-    pub fn metadata<P: AsRef<str>>(&self, path: P) -> io::Result<Metadata> {
+    pub async fn metadata<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<Metadata> {
         let path = from_utf8(path)?;
-        self.cap_std.metadata(path)
+        self.cap_std.metadata(path).await
     }
 
     /// Returns an iterator over the entries within `self`.
     #[inline]
-    pub fn entries(&self) -> io::Result<ReadDir> {
-        self.cap_std.entries().map(ReadDir::from_cap_std)
+    pub async fn entries(&self) -> io::Result<ReadDir> {
+        self.cap_std.entries().await.map(ReadDir::from_cap_std)
     }
 
     /// Returns an iterator over the entries within a directory.
@@ -201,9 +212,9 @@ impl Dir {
     /// This corresponds to [`async_std::fs::read_dir`], but only accesses
     /// paths relative to `self`.
     #[inline]
-    pub fn read_dir<P: AsRef<str>>(&self, path: P) -> io::Result<ReadDir> {
+    pub async fn read_dir<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<ReadDir> {
         let path = from_utf8(path)?;
-        self.cap_std.read_dir(path).map(ReadDir::from_cap_std)
+        self.cap_std.read_dir(path).await.map(ReadDir::from_cap_std)
     }
 
     /// Read the entire contents of a file into a bytes vector.
@@ -211,7 +222,7 @@ impl Dir {
     /// This corresponds to [`async_std::fs::read`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub async fn read<P: AsRef<str>>(&self, path: P) -> io::Result<Vec<u8>> {
+    pub async fn read<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<Vec<u8>> {
         let path = from_utf8(path)?;
         self.cap_std.read(path).await
     }
@@ -221,9 +232,9 @@ impl Dir {
     /// This corresponds to [`async_std::fs::read_link`], but only accesses
     /// paths relative to `self`.
     #[inline]
-    pub fn read_link<P: AsRef<str>>(&self, path: P) -> io::Result<String> {
+    pub async fn read_link<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<Utf8PathBuf> {
         let path = from_utf8(path)?;
-        self.cap_std.read_link(path).and_then(to_utf8)
+        self.cap_std.read_link(path).await.and_then(to_utf8)
     }
 
     /// Read the entire contents of a file into a string.
@@ -231,7 +242,7 @@ impl Dir {
     /// This corresponds to [`async_std::fs::read_to_string`], but only
     /// accesses paths relative to `self`.
     #[inline]
-    pub async fn read_to_string<P: AsRef<str>>(&self, path: P) -> io::Result<String> {
+    pub async fn read_to_string<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<String> {
         let path = from_utf8(path)?;
         self.cap_std.read_to_string(path).await
     }
@@ -241,9 +252,9 @@ impl Dir {
     /// This corresponds to [`async_std::fs::remove_dir`], but only accesses
     /// paths relative to `self`.
     #[inline]
-    pub fn remove_dir<P: AsRef<str>>(&self, path: P) -> io::Result<()> {
+    pub async fn remove_dir<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<()> {
         let path = from_utf8(path)?;
-        self.cap_std.remove_dir(path)
+        self.cap_std.remove_dir(path).await
     }
 
     /// Removes a directory at this path, after removing all its contents. Use
@@ -252,7 +263,7 @@ impl Dir {
     /// This corresponds to [`async_std::fs::remove_dir_all`], but only
     /// accesses paths relative to `self`.
     #[inline]
-    pub async fn remove_dir_all<P: AsRef<str>>(&self, path: P) -> io::Result<()> {
+    pub async fn remove_dir_all<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<()> {
         let path = from_utf8(path)?;
         self.cap_std.remove_dir_all(path).await
     }
@@ -263,8 +274,8 @@ impl Dir {
     /// as much as possible, removal is not guaranteed to be atomic with
     /// respect to a concurrent rename of the directory.
     #[inline]
-    pub fn remove_open_dir(self) -> io::Result<()> {
-        self.cap_std.remove_open_dir()
+    pub async fn remove_open_dir(self) -> io::Result<()> {
+        self.cap_std.remove_open_dir().await
     }
 
     /// Removes the directory referenced by `self`, after removing all its
@@ -274,8 +285,8 @@ impl Dir {
     /// as much as possible, removal is not guaranteed to be atomic with
     /// respect to a concurrent rename of the directory.
     #[inline]
-    pub fn remove_open_dir_all(self) -> io::Result<()> {
-        self.cap_std.remove_open_dir_all()
+    pub async fn remove_open_dir_all(self) -> io::Result<()> {
+        self.cap_std.remove_open_dir_all().await
     }
 
     /// Removes a file from a filesystem.
@@ -283,9 +294,9 @@ impl Dir {
     /// This corresponds to [`async_std::fs::remove_file`], but only accesses
     /// paths relative to `self`.
     #[inline]
-    pub fn remove_file<P: AsRef<str>>(&self, path: P) -> io::Result<()> {
+    pub async fn remove_file<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<()> {
         let path = from_utf8(path)?;
-        self.cap_std.remove_file(path)
+        self.cap_std.remove_file(path).await
     }
 
     /// Rename a file or directory to a new name, replacing the original file
@@ -294,7 +305,7 @@ impl Dir {
     /// This corresponds to [`async_std::fs::rename`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub fn rename<P: AsRef<str>, Q: AsRef<str>>(
+    pub async fn rename<P: AsRef<Utf8Path>, Q: AsRef<Utf8Path>>(
         &self,
         from: P,
         to_dir: &Self,
@@ -302,7 +313,7 @@ impl Dir {
     ) -> io::Result<()> {
         let from = from_utf8(from)?;
         let to = from_utf8(to)?;
-        self.cap_std.rename(from, &to_dir.cap_std, to)
+        self.cap_std.rename(from, &to_dir.cap_std, to).await
     }
 
     /// Changes the permissions found on a file or a directory.
@@ -311,9 +322,13 @@ impl Dir {
     /// accesses paths relative to `self`. Also, on some platforms, this
     /// function may fail if the file or directory cannot be opened for
     /// reading or writing first.
-    pub fn set_permissions<P: AsRef<str>>(&self, path: P, perm: Permissions) -> io::Result<()> {
+    pub async fn set_permissions<P: AsRef<Utf8Path>>(
+        &self,
+        path: P,
+        perm: Permissions,
+    ) -> io::Result<()> {
         let path = from_utf8(path)?;
-        self.cap_std.set_permissions(path, perm)
+        self.cap_std.set_permissions(path, perm).await
     }
 
     /// Query the metadata about a file without following symlinks.
@@ -321,9 +336,9 @@ impl Dir {
     /// This corresponds to [`async_std::fs::symlink_metadata`], but only
     /// accesses paths relative to `self`.
     #[inline]
-    pub fn symlink_metadata<P: AsRef<str>>(&self, path: P) -> io::Result<Metadata> {
+    pub async fn symlink_metadata<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<Metadata> {
         let path = from_utf8(path)?;
-        self.cap_std.symlink_metadata(path)
+        self.cap_std.symlink_metadata(path).await
     }
 
     /// Write a slice as the entire contents of a file.
@@ -331,7 +346,7 @@ impl Dir {
     /// This corresponds to [`async_std::fs::write`], but only accesses paths
     /// relative to `self`.
     #[inline]
-    pub async fn write<P: AsRef<str>, C: AsRef<[u8]>>(
+    pub async fn write<P: AsRef<Utf8Path>, C: AsRef<[u8]>>(
         &self,
         path: P,
         contents: C,
@@ -348,10 +363,14 @@ impl Dir {
     /// [`async_std::os::unix::fs::symlink`]: https://docs.rs/async-std/latest/async_std/os/unix/fs/fn.symlink.html
     #[cfg(not(windows))]
     #[inline]
-    pub fn symlink<P: AsRef<str>, Q: AsRef<str>>(&self, src: P, dst: Q) -> io::Result<()> {
+    pub async fn symlink<P: AsRef<Utf8Path>, Q: AsRef<Utf8Path>>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
         let src = from_utf8(src)?;
         let dst = from_utf8(dst)?;
-        self.cap_std.symlink(src, dst)
+        self.cap_std.symlink(src, dst).await
     }
 
     /// Creates a new file symbolic link on a filesystem.
@@ -362,10 +381,14 @@ impl Dir {
     /// [`async_std::os::windows::fs::symlink_file`]: https://docs.rs/async-std/latest/async_std/os/windows/fs/fn.symlink_file.html
     #[cfg(windows)]
     #[inline]
-    pub fn symlink_file<P: AsRef<str>, Q: AsRef<str>>(&self, src: P, dst: Q) -> io::Result<()> {
+    pub async fn symlink_file<P: AsRef<Utf8Path>, Q: AsRef<Utf8Path>>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
         let src = from_utf8(src)?;
         let dst = from_utf8(dst)?;
-        self.cap_std.symlink_file(src, dst)
+        self.cap_std.symlink_file(src, dst).await
     }
 
     /// Creates a new directory symlink on a filesystem.
@@ -376,10 +399,14 @@ impl Dir {
     /// [`async_std::os::windows::fs::symlink_dir`]: https://docs.rs/async-std/latest/async_std/os/windows/fs/fn.symlink_dir.html
     #[cfg(windows)]
     #[inline]
-    pub fn symlink_dir<P: AsRef<str>, Q: AsRef<str>>(&self, src: P, dst: Q) -> io::Result<()> {
+    pub async fn symlink_dir<P: AsRef<Utf8Path>, Q: AsRef<Utf8Path>>(
+        &self,
+        src: P,
+        dst: Q,
+    ) -> io::Result<()> {
         let src = from_utf8(src)?;
         let dst = from_utf8(dst)?;
-        self.cap_std.symlink_dir(src, dst)
+        self.cap_std.symlink_dir(src, dst).await
     }
 
     /// Creates a new `UnixListener` bound to the specified socket.
@@ -392,9 +419,12 @@ impl Dir {
     /// [`async_std::os::unix::net::UnixListener::bind`]: https://docs.rs/async-std/latest/async_std/os/unix/net/struct.UnixListener.html#method.bind
     #[cfg(unix)]
     #[inline]
-    pub fn bind_unix_listener<P: AsRef<str>>(&self, path: P) -> io::Result<UnixListener> {
+    pub async fn bind_unix_listener<P: AsRef<Utf8Path>>(
+        &self,
+        path: P,
+    ) -> io::Result<UnixListener> {
         let path = from_utf8(path)?;
-        self.cap_std.bind_unix_listener(path)
+        self.cap_std.bind_unix_listener(path).await
     }
 
     /// Connects to the socket named by path.
@@ -407,9 +437,9 @@ impl Dir {
     /// [`async_std::os::unix::net::UnixStream::connect`]: https://docs.rs/async-std/latest/async_std/os/unix/net/struct.UnixStream.html#method.connect
     #[cfg(unix)]
     #[inline]
-    pub fn connect_unix_stream<P: AsRef<str>>(&self, path: P) -> io::Result<UnixStream> {
+    pub async fn connect_unix_stream<P: AsRef<Utf8Path>>(&self, path: P) -> io::Result<UnixStream> {
         let path = from_utf8(path)?;
-        self.cap_std.connect_unix_stream(path)
+        self.cap_std.connect_unix_stream(path).await
     }
 
     /// Creates a Unix datagram socket bound to the given path.
@@ -422,9 +452,12 @@ impl Dir {
     /// [`async_std::os::unix::net::UnixDatagram::bind`]: https://docs.rs/async-std/latest/async_std/os/unix/net/struct.UnixDatagram.html#method.bind
     #[cfg(unix)]
     #[inline]
-    pub fn bind_unix_datagram<P: AsRef<str>>(&self, path: P) -> io::Result<UnixDatagram> {
+    pub async fn bind_unix_datagram<P: AsRef<Utf8Path>>(
+        &self,
+        path: P,
+    ) -> io::Result<UnixDatagram> {
         let path = from_utf8(path)?;
-        self.cap_std.bind_unix_datagram(path)
+        self.cap_std.bind_unix_datagram(path).await
     }
 
     /// Connects the socket to the specified address.
@@ -438,13 +471,15 @@ impl Dir {
     /// [`async_std::os::unix::net::UnixDatagram::connect`]: https://docs.rs/async-std/latest/async_std/os/unix/net/struct.UnixDatagram.html#method.connect
     #[cfg(unix)]
     #[inline]
-    pub fn connect_unix_datagram<P: AsRef<str>>(
+    pub async fn connect_unix_datagram<P: AsRef<Utf8Path>>(
         &self,
         unix_datagram: &UnixDatagram,
         path: P,
     ) -> io::Result<()> {
         let path = from_utf8(path)?;
-        self.cap_std.connect_unix_datagram(unix_datagram, path)
+        self.cap_std
+            .connect_unix_datagram(unix_datagram, path)
+            .await
     }
 
     /// Sends data on the socket to the specified address.
@@ -458,7 +493,7 @@ impl Dir {
     /// [`async_std::os::unix::net::UnixDatagram::send_to`]: https://docs.rs/async-std/latest/async_std/os/unix/net/struct.UnixDatagram.html#method.send_to
     #[cfg(unix)]
     #[inline]
-    pub fn send_to_unix_datagram_addr<P: AsRef<str>>(
+    pub async fn send_to_unix_datagram_addr<P: AsRef<Utf8Path>>(
         &self,
         unix_datagram: &UnixDatagram,
         buf: &[u8],
@@ -467,6 +502,7 @@ impl Dir {
         let path = from_utf8(path)?;
         self.cap_std
             .send_to_unix_datagram_addr(unix_datagram, buf, path)
+            .await
     }
 
     // async_std doesn't have `try_clone`.
@@ -476,9 +512,9 @@ impl Dir {
     /// This corresponds to [`async_std::path::Path::exists`], but only
     /// accesses paths relative to `self`.
     #[inline]
-    pub fn exists<P: AsRef<str>>(&self, path: P) -> bool {
+    pub async fn exists<P: AsRef<Utf8Path>>(&self, path: P) -> bool {
         match from_utf8(path) {
-            Ok(path) => self.cap_std.exists(path),
+            Ok(path) => self.cap_std.exists(path).await,
             Err(_) => false,
         }
     }
@@ -489,9 +525,9 @@ impl Dir {
     /// This corresponds to [`async_std::path::Path::is_file`], but only
     /// accesses paths relative to `self`.
     #[inline]
-    pub fn is_file<P: AsRef<str>>(&self, path: P) -> bool {
+    pub async fn is_file<P: AsRef<Utf8Path>>(&self, path: P) -> bool {
         match from_utf8(path) {
-            Ok(path) => self.cap_std.is_file(path),
+            Ok(path) => self.cap_std.is_file(path).await,
             Err(_) => false,
         }
     }
@@ -503,9 +539,9 @@ impl Dir {
     /// symbolic links to query information about the destination file. In case
     /// of broken symbolic links, this will return `false`.
     #[inline]
-    pub fn is_dir<P: AsRef<str>>(&self, path: P) -> bool {
+    pub async fn is_dir<P: AsRef<Utf8Path>>(&self, path: P) -> bool {
         match from_utf8(path) {
-            Ok(path) => self.cap_std.is_dir(path),
+            Ok(path) => self.cap_std.is_dir(path).await,
             Err(_) => false,
         }
     }
@@ -518,12 +554,47 @@ impl Dir {
     /// This function is not sandboxed and may access any path that the host
     /// process has access to.
     #[inline]
-    pub fn open_ambient_dir<P: AsRef<str>>(
+    pub async fn open_ambient_dir<P: AsRef<Utf8Path>>(
         path: P,
         ambient_authority: AmbientAuthority,
     ) -> io::Result<Self> {
         let path = from_utf8(path)?;
-        crate::fs::Dir::open_ambient_dir(path, ambient_authority).map(Self::from_cap_std)
+        crate::fs::Dir::open_ambient_dir(path, ambient_authority)
+            .await
+            .map(Self::from_cap_std)
+    }
+
+    /// Constructs a new instance of `Self` by opening the parent directory
+    /// (aka "..") of `self`, using the host process' ambient authority.
+    ///
+    /// # Ambient Authority
+    ///
+    /// This function accesses a directory outside of the `self` subtree.
+    #[inline]
+    pub async fn open_parent_dir<P: AsRef<Utf8Path>>(
+        &self,
+        ambient_authority: AmbientAuthority,
+    ) -> io::Result<Self> {
+        self.cap_std
+            .open_parent_dir(ambient_authority)
+            .await
+            .map(Self::from_cap_std)
+    }
+
+    /// Recursively create a directory and all of its parent components if they
+    /// are missing, using the host process' ambient authority.
+    ///
+    /// # Ambient Authority
+    ///
+    /// This function is not sandboxed and may access any path that the host
+    /// process has access to.
+    #[inline]
+    pub async fn create_ambient_dir_all<P: AsRef<Utf8Path>>(
+        path: P,
+        _ambient_authority: AmbientAuthority,
+    ) -> io::Result<()> {
+        let path = from_utf8(path)?;
+        fs::create_dir_all(path).await
     }
 }
 
