@@ -1,22 +1,20 @@
-use crate::{
-    fs::{Metadata, OpenOptions, Permissions},
-    fs_utf8::from_utf8,
-};
+use crate::fs::{Metadata, OpenOptions, Permissions};
+use crate::fs_utf8::from_utf8;
+use async_std::fs;
+use async_std::io::{self, IoSlice, IoSliceMut, Read, Seek, SeekFrom, Write};
 #[cfg(unix)]
 use async_std::os::unix::io::{AsRawFd, FromRawFd, IntoRawFd, RawFd};
 #[cfg(target_os = "wasi")]
 use async_std::os::wasi::io::{AsRawFd, FromRawFd, IntoRawFd, RawFd};
-use async_std::{
-    fs,
-    io::{self, IoSlice, IoSliceMut, Read, Seek, SeekFrom, Write},
-    task::{Context, Poll},
-};
+use async_std::task::{Context, Poll};
+use camino::Utf8Path;
 use cap_primitives::{ambient_authority, AmbientAuthority};
 #[cfg(not(windows))]
 use io_lifetimes::{AsFd, BorrowedFd, FromFd, IntoFd, OwnedFd};
 #[cfg(windows)]
 use io_lifetimes::{AsHandle, BorrowedHandle, FromHandle, IntoHandle, OwnedHandle};
-use std::{fmt, pin::Pin};
+use std::fmt;
+use std::pin::Pin;
 use unsafe_io::OwnsRaw;
 #[cfg(windows)]
 use {
@@ -118,15 +116,14 @@ impl File {
     /// This function is not sandboxed and may access any path that the host
     /// process has access to.
     #[inline]
-    pub fn open_ambient<P: AsRef<str>>(
+    pub async fn open_ambient<P: AsRef<Utf8Path>>(
         path: P,
         ambient_authority: AmbientAuthority,
     ) -> io::Result<Self> {
         let path = from_utf8(path)?;
-        Ok(Self::from_cap_std(crate::fs::File::open_ambient(
-            path,
-            ambient_authority,
-        )?))
+        crate::fs::File::open_ambient(path, ambient_authority)
+            .await
+            .map(Self::from_cap_std)
     }
 
     /// Constructs a new instance of `Self` with the options specified by
@@ -138,17 +135,15 @@ impl File {
     /// This function is not sandboxed and may access any path that the host
     /// process has access to.
     #[inline]
-    pub fn open_ambient_with<P: AsRef<str>>(
+    pub async fn open_ambient_with<P: AsRef<Utf8Path>>(
         path: P,
         options: &OpenOptions,
         ambient_authority: AmbientAuthority,
     ) -> io::Result<Self> {
         let path = from_utf8(path)?;
-        Ok(Self::from_cap_std(crate::fs::File::open_ambient_with(
-            path,
-            options,
-            ambient_authority,
-        )?))
+        crate::fs::File::open_ambient_with(path, options, ambient_authority)
+            .await
+            .map(Self::from_cap_std)
     }
 }
 
